//////////////////////////////////////////////////////////////////////////
//                                                                      //
// tanf.cpp -                                                          //
//                                                                      //
//                              //
//        .                          //
//                                                                      //
// Copyright (c) 2001 RC Module Inc.                                    //
//                                                                      //
//  :  ..                                                 //
//                                                                      //
//   $Revision:: 1     $      $Date:: 2/14/06 8:10p    $                //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#include "math.h"
#include "auxfd.h"

const int _iPI  = 0x3243F6A8;
const int _i2PI = 0x6487ED50;
const int _iPI2 = 0x1921FB54; //  Pi/2
const int _iPI4 = 0x0C90FDAA; //  Pi/4

//   
//  iTBL[i] = atan(1./pow(2,i)) * (1 << i) * 0x40000000
extern int _iTBL[42];
//          
extern int _iK[12];

const unsigned int _iPI_lowF  = 0x442d1847;   //     Pi
const unsigned int _iPI2_lowF = 0x22168c23;   //     Pi/2

#ifdef __cplusplus
extern "C" {
#endif

// tanf --
//
//    
//
//    --.
//    CORDIC,    .
float tanf(float f){

	int Mnt;               //   
	int Exp, Sgn, Spc;     //  ,     -  
	int NewExp;            //   
    int Arg;               //   

    //   float
	DestructFloat( f, Mnt, Exp, Sgn, Spc);

    //    > Pi\4,   
    bool bReverse = false;

	//  
	if ( Exp ==0x80 && Mnt==0 )	//  1 +inf
		return 0;

	if ( Exp == -0x7f && Mnt!=0 ){				//  3	Denormal
	    return f;
	}

	if ( Exp ==0x80 && Mnt!=0 )					//  4 NAN
		return f;

	if ( Exp > 26 ){			//    
		if (!_FPIEEE_Flags)
			_FPIEEE_Signal( _FPIEEE_TLoss ); 
		return ConstructFloat( 0, 0, 0, 0);
	}

	if ( Exp < -14 )			//    
		return f;			//	lim(x->0)(sin(x)) = x

    //      
	Arg = Mnt << 5;

    //  ,   
	
    unsigned int Arg_low=0;

    while (Exp>0 || !(Arg&0x30000000)){  //  n
		Exp--;
		Arg <<=1;
		Arg_low <<=1;
		if (Arg>_iPI){
			if (Arg_low>=_iPI_lowF){
				Arg -= _iPI;
				Arg_low -= _iPI_lowF;
			}
			else{
				Arg -= _iPI+1;
				Arg_low += 0x80000000 - _iPI_lowF;
			}
		}
		if (Arg_low & 0x80000000){
			Arg_low &= 0x7fffffff;
			Arg++;
		}
	}
	LONG tmp = Arg; //       
	if (Exp == 0){
		tmp = _iPI - tmp;		//   0 - n/2
		if (tmp<(_iPI2)){
			Arg = tmp;
			Sgn = 1 - Sgn;
			while (!(Arg&0x18000000)){
				Arg<<=1;
				Exp--;
			}
			Arg += _iPI_lowF >>(31+Exp);
			Arg -= Arg_low >>(31+Exp);
		}
		tmp = Arg;
	}
	if (Exp >= -1){
		Arg >>= -Exp;
		Exp = 0;
		tmp = _iPI2 - Arg;		//   0 - n/4
		if ( tmp < _iPI4 ){
			Arg = tmp;
			bReverse = !bReverse;
			while (!(Arg&0x18000000)){
				Arg<<=1;
				Exp--;
			}
			Arg += _iPI2_lowF >>(31+Exp);
			Arg -= Arg_low >>(31+Exp);
		}
		Arg <<=1;
		Exp--;
	}

	Arg = Arg << 1;

	LONG y = 0; 
	NewExp = 0;

	//	 , , CORDIC
    //   SIN, COS

    if (Exp == 0){
		NewExp=0;
	}
	else{
		NewExp= Exp+1;
		Arg >>= 1;
	}
	int i = -NewExp;
	int x = _iK[i];
	int x1 = x;

    while (i<32+NewExp){  //  
		if (Arg >= 0){
			x1 -= y >> (i-NewExp);
			y += x >> (i+NewExp);
			Arg = (Arg - _iTBL[i++])<<1;
		}
		else{
			x1 += y >> (i-NewExp);
			y -= x >> (i+NewExp);
			Arg = (Arg + _iTBL[i++])<<1;
		}
		x = x1;
	}
	while (i<32-NewExp){  //  
		if (Arg >= 0){
			y += x >> (i+NewExp);
			Arg = (Arg - _iTBL[i++])<<1;
		}
		else{
			y -= x >> (i+NewExp);
			Arg = (Arg + _iTBL[i++])<<1;
		}
	}
	if (bReverse){

		x = y;
		y = x1;
		NewExp = -NewExp;
	}

    //   TAN = SIN/COS
    //	y = _DivFix0x64(y, x, NewExp);

	i=0;
    x1=y;
    y=0;
	while (x<=x1){
		NewExp++;
		x<<=1;
	}
	NewExp--;
	x>>=1;
	while(i<30){
		tmp = x1;
		tmp -= x;
		if (tmp>=0){
			x1 = tmp;
			y++;
		}
		x1<<=1;
		y<<=1;
		i++;
	}

    //   
	if (y!=0){
		while (!(y & 0x40000000)){
			y <<= 1;
			NewExp--;
		}
	}
	else
		NewExp = -0x7f;

	y >>=6;	//    , ...
	y = (y>>1) + (y & 1); //   ,  .

	if (y == 0x01000000){	//				!!!!!!!!!!
		y = 0;
		NewExp++;
	}
    //   float
	return ConstructFloat( y, NewExp, Sgn, 0);		
}

#ifdef __cplusplus
}
#endif
